"""
Simple run result formatters
"""

from ...project import Stage
from ...steps.models import Artifact, ArtifactType
from ...steps.run import RunResult
from ...utilities.junit import to_test_suites, sum_suites


def to_string(run_result: RunResult) -> str:
    result: str = ""
    for stage in Stage:
        run_results = run_result.results_for_stage(stage)
        if run_results:
            result += f"Stage {stage.name}\n"
            for res in run_results:
                result += f"{res.timestamp} - {res.project.name} - {res.stage} - success: {res.output.success} \n"
                artifact = res.output.produced_artifact
                if artifact and artifact.artifact_type == ArtifactType.JUNIT_TESTS:
                    result += to_test_report(artifact)

    return result


def to_test_report(artifact: Artifact) -> str:
    """ Gather the first test suites and test cases. Output is truncated to around 200 lines to not overwhelm Github """
    test_result = []
    suites = to_test_suites(artifact)
    total_tests = sum_suites(suites)
    test_result.append(f"{total_tests} \n\n")
    for suite in suites:
        test_result.append(f"Suite {suite.name}\n")
        if len(test_result) > 200:
            test_result.append("  ... output truncated\n")
            break
        for case in suite:
            test_result.append(f"  Case {case.name} \n")
    return "".join(test_result)



""" Step that tests the docker image from the target `tester` in Dockerfile-mpl.


## 🧪 Testing inside a container

When unit tests are run within a docker container the test results need to be written to a folder inside it.
This means that the test step _within the docker container_ should not return a system error.
Otherwise, building of the container would stop and the test results would not be committed to a layer.

The test results need to be writted  written to a folder named `$WORKDIR/target/test-reports/` for
`TestDocker.extract_test_results` to find and extract them.


"""
from logging import Logger

from python_on_whales import Container

from .after_test import IntegrationTestAfter
from .before_test import IntegrationTestBefore
from .. import Step, Meta
from ..models import Input, Output, ArtifactType, input_to_artifact, Artifact
from ...project import Stage, Project
from ...utilities.docker import DockerConfig, build, docker_image_tag, docker_file_path, docker_copy, \
    remove_container, create_container
from ...utilities.junit import to_test_suites, sum_suites, TEST_OUTPUT_PATH_KEY, TEST_RESULTS_URL_KEY


class TestDocker(Step):
    def __init__(self, logger: Logger) -> None:
        meta = Meta(name='Docker Test', description='Test docker image', version='0.0.1', stage=Stage.TEST)
        super().__init__(
            logger=logger, meta=meta,
            produced_artifact=ArtifactType.JUNIT_TESTS,
            required_artifact=ArtifactType.NONE,
            before=IntegrationTestBefore(logger),
            after=IntegrationTestAfter(logger)
        )

    def execute(self, step_input: Input) -> Output:
        docker_config = DockerConfig.from_dict(step_input.run_properties.config)
        test_target = docker_config.test_target
        if not test_target:
            raise ValueError('docker.testTarget must be specified')

        tag = docker_image_tag(step_input) + '-test'
        project = step_input.project
        dockerfile = docker_file_path(project=project, docker_config=docker_config)
        success = build(logger=self._logger, root_path=docker_config.root_folder,
                        file_path=dockerfile, image_tag=tag, target=test_target)
        container = create_container(self._logger, tag)

        if success:
            artifact = self.extract_test_results(self._logger, project, container, step_input)

            suite = to_test_suites(artifact)
            summary = sum_suites(suite)

            output = Output(success=summary.is_success,
                            message=f"Tests results produced for {project.name} ({summary})",
                            produced_artifact=artifact)
        else:
            output = Output(success=False,
                            message=f"Tests failed to run for {project.name}. No test results have been recorded.",
                            produced_artifact=None)

        remove_container(self._logger, container)

        return output

    @staticmethod
    def extract_test_results(logger: Logger, project: Project, container: Container, step_input: Input) -> Artifact:
        path_in_container = f'{project.test_report_path}/.'
        docker_copy(logger=logger, container_path=path_in_container, dst_path=project.test_report_path,
                    container=container)

        return input_to_artifact(artifact_type=ArtifactType.JUNIT_TESTS, step_input=step_input,
                                 spec={TEST_OUTPUT_PATH_KEY: project.test_report_path,
                                       TEST_RESULTS_URL_KEY: step_input.run_properties.details.tests_url})



"""Commands related to build"""
import asyncio
import shutil
from pathlib import Path
from typing import Optional

import click
import questionary
from click import ParamType, BadParameter
from click.shell_completion import CompletionItem
from github import Github
from github.GitRelease import GitRelease
from questionary import Choice
from rich.console import Console
from rich.markdown import Markdown

from . import CliContext, CONFIG_PATH_HELP, check_updates, get_meta_version
from . import create_console_logger
from .commands.build.jenkins import JenkinsRunParameters, run_jenkins, get_token
from .commands.build.mpyl import MpylRunParameters, run_mpyl, MpylCliParameters, MpylRunConfig, find_build_set
from ..constants import DEFAULT_CONFIG_FILE_NAME, DEFAULT_RUN_PROPERTIES_FILE_NAME, BUILD_ARTIFACTS_FOLDER
from ..project import load_project
from ..reporting.formatting.markdown import run_result_to_markdown
from ..steps.models import RunProperties
from ..steps.run import RunResult
from ..utilities.github import GithubConfig
from ..utilities.pyaml_env import parse_config
from ..utilities.repo import Repository, RepoConfig


async def warn_if_update(console: Console):
    version = get_meta_version()
    update = await check_updates(meta=version)
    if update:
        console.print(Markdown(f"⚠️  **You can upgrade from {version} to {update} :** `pip install -U mpyl=={update}`"))


@click.group('build')
@click.option('--config', '-c', required=True, type=click.Path(exists=True), help=CONFIG_PATH_HELP,
              envvar="MPYL_CONFIG_PATH", default=DEFAULT_CONFIG_FILE_NAME)
@click.option('--properties', '-p', required=False, type=click.Path(exists=False), help='Path to run properties',
              envvar="MPYL_RUN_PROPERTIES_PATH", default=DEFAULT_RUN_PROPERTIES_FILE_NAME, show_default=True)
@click.option('--verbose', '-v', is_flag=True, default=False, show_default=True, help='Verbose output')
@click.pass_context
def build(ctx, config, properties, verbose):
    """Pipeline build commands"""
    console = create_console_logger(local=False, verbose=verbose)
    parsed_config = parse_config(config)
    parsed_properties = parse_config(properties)

    repo = ctx.with_resource(Repository(config=RepoConfig.from_config(parsed_config)))
    ctx.obj = CliContext(parsed_config, repo, console, verbose, parsed_properties)


@build.command(help='Run an MPyL build')
@click.option('--ci', is_flag=True,
              help='Run as CI build instead of local. Ignores unversioned changes.')
@click.option('--all', 'all_', is_flag=True, help='Build all projects, regardless of changes on branch')
@click.option(
    '--tag', '-t',
    help='Tag to build',
    type=click.STRING,
    required=False
)
@click.pass_obj
def run(obj: CliContext, ci, all_, tag):  # pylint: disable=invalid-name
    asyncio.run(warn_if_update(obj.console))
    run_properties = RunProperties.from_configuration(obj.run_properties, obj.config) if ci \
        else RunProperties.for_local_run(obj.config, obj.repo.get_sha, obj.repo.get_branch, tag)

    parameters = MpylCliParameters(local=not ci, pull_main=all_, all=all_, verbose=obj.verbose, tag=tag,
                                   target=run_properties.target)
    obj.console.log(parameters)
    run_parameters = MpylRunParameters(
        run_config=MpylRunConfig(config=obj.config, run_properties=run_properties),
        parameters=parameters
    )
    run_mpyl(run_parameters, None)


@build.command(help="The status of the current local branch from MPyL's perspective")
@click.pass_obj
def status(obj: CliContext):
    try:
        upgrade_check = asyncio.wait_for(warn_if_update(obj.console), timeout=3)
        __print_status(obj)
        asyncio.get_event_loop().run_until_complete(upgrade_check)
    except asyncio.exceptions.TimeoutError:
        pass


def __print_status(obj: CliContext):
    run_properties = RunProperties.from_configuration(obj.run_properties, obj.config)
    branch = run_properties.versioning.branch
    if not branch:
        branch = obj.repo.get_branch
        obj.console.log(
            f'Branch not specified in `{DEFAULT_RUN_PROPERTIES_FILE_NAME}`. Branch determined via git: {branch}')

    if branch and obj.repo.main_branch == obj.repo.get_branch:
        obj.console.log(f'On main branch ({branch}), cannot determine build status')
        return
    tag = obj.repo.get_tag if not branch else None
    version = run_properties.versioning
    revision = "Tag" if tag else "Branch"
    obj.console.print(Markdown(f"**{revision}:** `{version.branch or version.tag}` at `{version.revision}`"))

    if obj.repo.main_branch_pulled:
        changes = obj.repo.changes_in_branch_including_local() if branch else obj.repo.changes_in_merge_commit()
        build_set = find_build_set(obj.repo, changes, False)
        result = RunResult(run_properties=run_properties, run_plan=build_set)
        if result.run_plan:
            obj.console.print(Markdown("**Execution plan:**  \n" + run_result_to_markdown(result)))
        else:
            obj.console.print("No changes detected, nothing to do.")


class Pipeline(ParamType):
    name = 'pipeline'

    def shell_complete(self, ctx: click.Context, param, incomplete: str):
        if not ctx.parent or ctx.parent.params['config'] is None or not Path(ctx.parent.params['config']).exists():
            raise BadParameter('Either --config parameter must or MPYL_CONFIG_PATH env var must be set', ctx=ctx,
                               param=param)

        config: dict = parse_config(ctx.parent.params['config'])
        parsed_config: dict[str, str] = config['jenkins']['pipelines']

        return [
            CompletionItem(value=pl[0], help=pl[1]) for pl in parsed_config.items() if
            incomplete in pl[0]
        ]


def select_tag(ctx) -> str:
    console = Console()
    with console.status("Fetching tags...") as spinner:
        github_config = GithubConfig.from_config(ctx.obj.config)
        github = Github(login_or_token=get_token(github_config))

        repo = github.get_repo(github_config.repository)

        def to_choice(git_release: GitRelease):
            title = git_release.title + " " + git_release.body.split('* ')[-1].splitlines()[0]
            return Choice(title=title, value=git_release.title)

        choices = map(to_choice, repo.get_releases())
        user_name = github.get_user().login

        def by_choice(choice: Choice):
            # prioritize own tags
            if user_name in str(choice.title):
                return f"9{choice.value}"
            return choice.value

        sorted_choices = sorted(choices, key=by_choice, reverse=True)

        spinner.stop()
        release_id = questionary.select("Which tag do you want to release?", show_selected=True,
                                        choices=sorted_choices).ask()
        release = repo.get_release(release_id)
        return release.tag_name


def ask_for_input(ctx, _param, value) -> Optional[str]:
    if value == "not_set":
        return None
    if value == "prompt":
        return select_tag(ctx)
    return value


@build.command(help='Run a multi branch pipeline build on Jenkins')
@click.option(
    '--user', '-u',
    help='Authentication API user. Can be set via env var JENKINS_USER',
    envvar="JENKINS_USER",
    type=click.STRING,
    required=True
)
@click.option(
    '--password', '-p',
    help='Authentication API password. Can be set via env var JENKINS_PASSWORD',
    envvar="JENKINS_PASSWORD",
    type=click.STRING,
    required=True
)
@click.option(
    '--pipeline', '-pl',
    help='The pipeline to run. Must be one of the pipelines listed in `jenkins.pipelines`. '
         'Default value is `jenkins.defaultPipeline`',
    type=Pipeline(),
    required=False
)
@click.option(
    '--test', '-t',
    help='A specific test version on https://test.pypi.org/project/mpyl/ to use for the build.',
    type=click.STRING,
    required=False
)
@click.option(
    '--arguments', '-a',
    multiple=True,
    help='A series of arguments to pass to the pipeline. Note that will run within the pipenv in jenkins. '
         'To execute `mpyl build status`, pass `-a run -a mpyl -a build -a status`',
)
@click.option(
    '--background', '-bg',
    help="Starts Jenkins build in a 'fire and forget' fashion. "
         "Can be set via env var MPYL_JENKINS_BACKGROUND",
    envvar="MPYL_JENKINS_BACKGROUND",
    is_flag=True,
    default=False
)
@click.option(
    '--silent', '-s',
    help="Indicates whether to show Jenkins' logging or not. "
         "Can be set via env var MPYL_JENKINS_SILENT",
    envvar="MPYL_JENKINS_SILENT",
    is_flag=True,
    default=False
)
@click.option('--tag', '-tg', is_flag=False, flag_value="prompt", default="not_set", callback=ask_for_input)
@click.pass_context
def jenkins(ctx, user, password, pipeline, test, arguments, background, silent,  # pylint: disable=too-many-arguments
            tag):
    try:
        upgrade_check = asyncio.wait_for(warn_if_update(ctx.obj.console), timeout=5)
        selected_pipeline = pipeline if pipeline else ctx.obj.config['jenkins']['defaultPipeline']
        pipeline_parameters = {'TEST': 'true', 'VERSION': test} if test else {}
        if arguments:
            pipeline_parameters['BUILD_PARAMS'] = " ".join(arguments)

        run_argument = JenkinsRunParameters(
            jenkins_user=user, jenkins_password=password, config=ctx.obj.config,
            pipeline=selected_pipeline, pipeline_parameters=pipeline_parameters,
            verbose=not silent or ctx.obj.verbose,
            follow=not background, tag=tag
        )

        run_jenkins(run_argument)
        asyncio.get_event_loop().run_until_complete(upgrade_check)
    except asyncio.exceptions.TimeoutError:
        pass


@build.command(help=f'Clean MPyL metadata in `{BUILD_ARTIFACTS_FOLDER}` folders')
@click.option('--filter', '-f', 'filter_', required=False, type=click.STRING, help='Filter based on filepath ')
@click.pass_obj
def clean(obj: CliContext, filter_):
    found_projects: list[Path] = [
        Path(load_project(obj.repo.root_dir(), Path(project_path), strict=False).target_path)
        for project_path in obj.repo.find_projects(filter_ if filter_ else '')
    ]

    paths_to_clean = [path for path in found_projects if path.exists()]
    if paths_to_clean:
        for target_path in paths_to_clean:
            shutil.rmtree(target_path)
            obj.console.print(f"🧹 Cleaned up {target_path}")
    else:
        obj.console.print("Nothing to clean")


if __name__ == '__main__':
    build()  # pylint: disable=no-value-for-parameter



"""Docker related utility methods"""
import logging
import shlex

from dataclasses import dataclass
from itertools import tee
from logging import Logger
from traceback import print_exc
from typing import Dict, Optional, Iterator, cast, Union
import shutil
from pathlib import Path
from python_on_whales import docker, Image, Container, DockerException
from python_on_whales.exceptions import NoSuchContainer
from rich.text import Text

from ..logging import try_parse_ansi
from ...project import Project
from ...steps.models import Input


@dataclass(frozen=True)
class DockerComposeConfig:
    period_seconds: int
    failure_threshold: int

    @property
    def total_duration(self):
        return self.period_seconds * self.failure_threshold

    @staticmethod
    def from_yaml(config: dict):
        compose_config = config.get('docker', {}).get('compose')
        if not compose_config:
            raise KeyError('docker.compose needs to be defined')
        return DockerComposeConfig(period_seconds=int(compose_config['periodSeconds']),
                                   failure_threshold=int(compose_config['failureThreshold']))


@dataclass(frozen=True)
class DockerConfig:
    host_name: str
    user_name: str
    password: str
    root_folder: str
    build_target: Optional[str]
    test_target: Optional[str]
    docker_file_name: str

    @staticmethod
    def from_dict(config: Dict):
        try:
            registry: Dict = config['docker']['registry']
            build_config: Dict = config['docker']['build']
            return DockerConfig(
                host_name=registry['hostName'],
                user_name=registry['userName'],
                password=registry['password'],
                root_folder=build_config['rootFolder'],
                build_target=build_config.get('buildTarget', None),
                test_target=build_config.get('testTarget', None),
                docker_file_name=build_config['dockerFileName']
            )
        except KeyError as exc:
            raise KeyError(f'Docker config could not be loaded from {config}') from exc


def execute_with_stream(logger: Logger, container: Container, command: str, task_name: str):
    result = cast(Iterator[tuple[str, bytes]], container.execute(command=shlex.split(command), stream=True))
    result_list = stream_docker_logging(logger, result, task_name)

    return result_list


def stream_docker_logging(logger: Logger, generator: Union[Iterator[str], Iterator[tuple[str, bytes]]], task_name: str,
                          level=logging.INFO) -> list[str]:
    copied_logs = []

    while True:
        try:
            next_item = next(generator)
            log_line = next_item[1].decode(errors="replace") if isinstance(next_item, tuple) else next_item
            copied_logs.append(log_line)
            logger.log(level, try_parse_ansi(log_line))
        except StopIteration:
            logger.info(f'{task_name} complete.')
            return copied_logs


def docker_image_tag(step_input: Input):
    git = step_input.run_properties.versioning
    tag = git.tag if git.tag else f"pr-{git.pr_number}"
    return f"{step_input.project.name.lower()}:{tag}".replace('/', '_')


def docker_file_path(project: Project, docker_config: DockerConfig):
    return f'{project.deployment_path}/{docker_config.docker_file_name}'


def docker_copy(logger: Logger, container_path: str, dst_path: str, container: Container):
    """
    Copies the contents of the specified path within the container to a locally created destination

    :param logger: the logger
    :param container_path: the path of the directory in the container to copy
    :param dst_path: the path to copy the container content to
    :param container: the container to copy from
    """
    shutil.rmtree(dst_path, ignore_errors=True)
    Path(dst_path).mkdir(parents=True, exist_ok=True)

    if not docker.container.exists(container.id):
        raise ValueError(f'Container {container.id} does not exist')

    logger.info(
        f"Copying contents from container {container.id} at "
        f"path {container_path} to host at {dst_path}"
    )
    try:
        docker.copy(f'{container.id}:{container_path}', dst_path)
    except NoSuchContainer as exc:
        logger.warning(f'Could not find data in container {container.name} at expected location {container_path}')
        raise exc


def build(logger: Logger, root_path: str, file_path: str, image_tag: str, target: str) -> bool:
    """
    :param logger: the logger
    :param root_path: the root path to which `docker_file_path` is relative
    :param file_path: path to the docker file to be built
    :param image_tag: the tag of the image
    :param target: the 'target' within the multi-stage docker image
    :return: True if success, False if failure
    """
    logger.info(f"Building docker image with {file_path} and target {target}")

    try:
        logs = docker.buildx.build(context_path=root_path, file=file_path, tags=[image_tag], target=target,
                                   stream_logs=True)
        if logs is not None and not isinstance(logs, Image):
            stream_docker_logging(logger=logger, generator=logs, task_name=f'Build {file_path}:{target}')
        logger.debug(logs)
        return True

    except DockerException as exc:
        command = " ".join(exc.docker_command)
        logger.warning(f"Docker build failed with command {command} and exit code {exc.return_code}")
        return False
    except Exception as exc:  # pylint: disable=broad-exception-caught
        print(f"Docker build failed with {exc.__class__.__name__}")
        print_exc()
        return False


def login(logger: Logger, docker_config: DockerConfig) -> None:
    logger.info(f"Logging in with user '{docker_config.user_name}'")
    docker.login(server=f'https://{docker_config.host_name}', username=docker_config.user_name,
                 password=docker_config.password)
    logger.debug(f"Logged in as '{docker_config.user_name}'")


def create_container(logger: Logger, image_name: str) -> Container:
    logger.info(f"Creating container from image {image_name}")
    container = docker.create(image_name)
    logger.info(f"Created container {container.id}")

    return container


def remove_container(logger: Logger, container: Container) -> None:
    logger.info(f"Removing container {container.id}")
    docker.remove(container.id)
    logger.info(f"Removed container {container.id}")



"""Console logging for utilities for the `rich` library."""

from rich.errors import ConsoleError
from rich.text import Text


def try_parse_ansi(text: str):
    try:
        return Text.from_ansi(text)
    except ConsoleError:
        return Text(text)



""" Step that runs relevant cypress tests in the post deploy stage """

import os

from logging import Logger

from python_on_whales import docker, Container, DockerException

from .. import Step, Meta
from ..models import ArtifactType, Input, Output, input_to_artifact
from ...project import Stage, Target
from ...utilities.cypress import CypressConfig
from ...utilities.docker import execute_with_stream
from ...utilities.junit import TEST_OUTPUT_PATH_KEY, TEST_RESULTS_URL_KEY


class CypressTest(Step):
    def __init__(self, logger: Logger) -> None:
        super().__init__(logger, Meta(
            name='Cypress Test',
            description='Step to run cypress tests',
            version='0.0.1',
            stage=Stage.POST_DEPLOY
        ), produced_artifact=ArtifactType.JUNIT_TESTS, required_artifact=ArtifactType.NONE)

    def execute(self, step_input: Input) -> Output:
        if step_input.run_properties.target == Target.PRODUCTION:
            return Output(success=True, message="Cypress tests are not run on production")

        self._logger.info(f"Running cypress tests for project {step_input.project.name}")
        cypress_config = CypressConfig.from_config(step_input.run_properties.config)
        volume_path = os.path.join(os.getcwd(), cypress_config.cypress_source_code_path)

        if step_input.project.dependencies and step_input.project.dependencies.postdeploy:
            specs_string = ', '.join(step_input.project.dependencies.postdeploy)
        else:
            raise ValueError("No cypress specs are defined in the project dependencies")

        custom_image_tag = "mpyl/cypress"
        docker.build(context_path=volume_path, tags=[custom_image_tag], file=f"{volume_path}/Dockerfile-mpyl")
        docker_container = docker.run(image=custom_image_tag, interactive=True, detach=True,
                                      volumes=[
                                          (volume_path, "/cypress"),
                                          (os.path.expanduser(cypress_config.kubectl_config_path), "/root/.kube/config")
                                      ],
                                      workdir="/cypress")
        if not isinstance(docker_container, Container):
            raise TypeError("Docker run command should return a container")

        reports_folder = f"reports/{step_input.project.name}"
        artifact = input_to_artifact(artifact_type=ArtifactType.JUNIT_TESTS, step_input=step_input,
                                     spec={TEST_OUTPUT_PATH_KEY: f"{volume_path}/{reports_folder}",
                                           TEST_RESULTS_URL_KEY: ''})

        try:
            execute_with_stream(logger=self._logger, container=docker_container,
                                command='bash -c "cp cypress.env.json.example cypress.env.json && '
                                        f"sed -i 's/acceptance/"
                                        f"{CypressTest._target_to_test_target(step_input.run_properties.target)}"
                                        f"/' cypress.env.json && "
                                        f"sed -i 's/{{PR_NUMBER}}/{step_input.run_properties.versioning.pr_number}/' "
                                        'cypress.env.json"',
                                task_name="Preparing env file")
            execute_with_stream(logger=self._logger, container=docker_container, command="yarn install",
                                task_name="Running yarn install")
            execute_with_stream(logger=self._logger, container=docker_container, command="yarn cypress install",
                                task_name="Installing cypress")
            execute_with_stream(logger=self._logger, container=docker_container, command="yarn cypress verify",
                                task_name="Verifying cypress")
            execute_with_stream(logger=self._logger, container=docker_container, command="yarn tsc",
                                task_name="Compiling typescript")
            execute_with_stream(logger=self._logger, container=docker_container, command=f"rm -rf {reports_folder}",
                                task_name="Remove old report files")

            run_command = f'bash -c "yarn cypress run --spec {specs_string} --reporter-options mochaFile=' \
                          f'"{reports_folder}/[hash].xml" || true"'
            record_key = cypress_config.record_key
            if record_key:
                run_command = f'bash -c "yarn cypress run --spec {specs_string} --reporter-options ' \
                              f'"mochaFile={reports_folder}/[hash].xml" --record --key ' \
                              f'b6a2aab1-0b80-4ca0-a56c-1c8d98a8189c || true "'
            result = execute_with_stream(logger=self._logger, container=docker_container, command=run_command,
                                         task_name="Running cypress tests")

            for stdout in result:
                if record_key and "Recorded Run" in stdout:
                    artifact.spec[TEST_RESULTS_URL_KEY] = stdout.rstrip().rsplit('Recorded Run: ', 1)[1]
                if "error Command failed with exit code" in stdout:
                    raise DockerException(command_launched=[run_command], return_code=1)
        except DockerException:
            return Output(success=False,
                          message=f"Cypress tests for project {step_input.project.name} have one or more failures",
                          produced_artifact=artifact)
        finally:
            docker_container.stop()
            docker_container.remove()

        return Output(success=True, message=f"Cypress tests for project {step_input.project.name} passed",
                      produced_artifact=artifact)

    @staticmethod
    def _target_to_test_target(target: Target) -> str:
        if target == Target.PULL_REQUEST_BASE:
            return 'test'
        if target == Target.ACCEPTANCE:
            return 'acceptance'

        return 'pr'



"""Kubernetes deployment related helper methods"""
from logging import Logger
from typing import Optional

from kubernetes import config, client

from ...deploy.k8s.resources import CustomResourceDefinition
from ...models import RunProperties
from ....project import Project, Target, ProjectName
from ....steps import Input, Output
from ....steps.deploy.k8s import helm
from ....steps.deploy.k8s.rancher import cluster_config, rancher_namespace_metadata, ClusterConfig


def get_namespace(run_properties: RunProperties, project: Project) -> Optional[str]:
    if run_properties.target == Target.PULL_REQUEST:
        return run_properties.versioning.identifier

    return get_namespace_from_project(project)


def get_namespace_from_project(project: Project) -> Optional[str]:
    if project.deployment and project.deployment.namespace:
        return project.deployment.namespace

    return None


def upsert_namespace(logger: Logger, step_input: Input, context: str):
    properties = step_input.run_properties

    config.load_kube_config(context=context)
    logger.info(f"Deploying target {properties.target} and k8s context {context}")
    api = client.CoreV1Api()

    namespace = get_namespace(properties, step_input.project)
    meta_data = rancher_namespace_metadata(namespace or step_input.project.name, step_input)
    namespaces = api.list_namespace(field_selector=f'metadata.name={namespace}')

    if len(namespaces.items) == 0 and not step_input.dry_run:
        api.create_namespace(client.V1Namespace(api_version='v1', kind='Namespace', metadata=meta_data))
    else:
        logger.info(f"Found namespace {namespace}")

    return namespace


def deploy_helm_chart(logger: Logger, chart: dict[str, CustomResourceDefinition], step_input: Input,
                      target_cluster: ClusterConfig,
                      release_name: str, delete_existing: bool = False) -> Output:
    namespace = upsert_namespace(logger, step_input, target_cluster.context)

    return helm.install(logger, chart, step_input, release_name, namespace, target_cluster.context, delete_existing)


def substitute_namespaces(env_vars: dict[str, str], all_projects: set[ProjectName],
                          projects_to_deploy: set[ProjectName],
                          pr_identifier: Optional[int]) -> dict[str, str]:
    env = env_vars.copy()

    def get_namespace_for_linked_project(project_name: ProjectName):
        is_part_of_same_deploy_set = project_name in projects_to_deploy
        if is_part_of_same_deploy_set and pr_identifier:
            return f'pr-{pr_identifier}'
        return project_name.namespace

    def replace_namespace(env_value, project_name, namespace):
        search_value = project_name + '.{namespace}'
        replace_value = project_name + '.' + namespace
        return env_value.replace(search_value, replace_value)

    for project in all_projects:
        if project.namespace:
            linked_project_namespace = get_namespace_for_linked_project(project)
            for key, value in env.items():
                replaced_namespace = replace_namespace(value, project.name, linked_project_namespace)
                updated_pr = replaced_namespace.replace('{PR-NUMBER}',
                                                        str(pr_identifier)) if pr_identifier else replaced_namespace
                env[key] = updated_pr
    return env



""" Defines information about the repository, any changes made to it and the containing projects.
`mpyl.utilities.repo.Repository` is a facade for the Version Control System.
At this moment Git is the only supported VCS.
"""

import logging
from dataclasses import dataclass
from pathlib import Path
from typing import Dict, Optional
from urllib.parse import urlparse

from git import Git, Repo, Remote
from git.objects import Commit

from ...project import Project


@dataclass(frozen=True)
class Revision:
    ord: int
    """Ordinal number indicating how this revision ranks historically"""
    hash: str
    """Git hash for this revision"""
    files_touched: set[str]
    """Paths to files that were altered in this hash"""


@dataclass(frozen=True)
class RepoCredentials:
    url: str
    user_name: str
    password: str

    @property
    def to_url_with_credentials(self):
        parsed = urlparse(self.url)
        return f"{parsed.scheme}://{self.user_name}:{self.password}@{parsed.netloc}{parsed.path}"

    @staticmethod
    def from_config(config: Dict):
        return RepoCredentials(url=config['url'], user_name=config['userName'], password=config['password'])


@dataclass(frozen=True)
class RepoConfig:
    main_branch: str
    ignore_patterns: list[str]
    repo_credentials: Optional[RepoCredentials]

    @staticmethod
    def from_config(config: Dict):
        git_config = config['cvs']['git']
        maybe_remote_config = git_config.get('remote', None)
        return RepoConfig(
            main_branch=git_config['mainBranch'],
            ignore_patterns=git_config.get('ignorePatterns', []),
            repo_credentials=RepoCredentials.from_config(maybe_remote_config) if maybe_remote_config else None
        )


class Repository:

    def __init__(self, config: RepoConfig):
        self._config = config
        self._root_dir = Git().rev_parse('--show-toplevel')
        self._repo = Repo(self._root_dir)  # pylint: disable=attribute-defined-outside-init

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._repo.close()
        return self

    @property
    def get_sha(self):
        return self._repo.head.commit.hexsha

    @property
    def get_short_sha(self):
        return self._repo.git.rev_parse(self._repo.head, short=True)

    @property
    def get_branch(self) -> Optional[str]:
        if self._repo.head.is_detached:
            return None
        return self._repo.active_branch.name

    @property
    def get_tag(self) -> Optional[str]:
        current_revision = self._repo.head.commit
        current_tag = self._repo.git.describe(current_revision, tags=True)
        logging.debug(f"Current revision: {current_revision} tag: {current_tag}")
        return current_tag

    @property
    def get_remote_url(self):
        return self._repo.remote().url

    def root_dir(self) -> Path:
        return Path(self._root_dir)

    @property
    def main_branch(self) -> str:
        return self._config.main_branch

    def __get_filter_patterns(self):
        return ["--"] + [f":!{pattern}" for pattern in self._config.ignore_patterns]

    @property
    def latest_tag(self) -> str:
        return str(sorted(self._repo.tags, key=lambda t: t.commit.committed_datetime)[-1])

    def __to_revision(self, count: int, revision: Commit, files_touched_in_branch: set[str]) -> Revision:
        files_in_revision = set(
            self._repo.git.diff_tree(self.__get_filter_patterns(), no_commit_id=True, name_only=True,
                                     r=str(revision)).splitlines())
        intersection = files_in_revision.intersection(files_touched_in_branch)
        return Revision(count, str(revision), intersection)

    def changes_in_branch(self) -> list[Revision]:
        revisions = list(reversed(list(self._repo.iter_commits(f"{self._config.main_branch}..HEAD"))))
        files_touched_in_branch = set(
            self._repo.git.diff(f'{revisions[0].hexsha}..{revisions[-1].hexsha}', name_only=True).splitlines())
        return [self.__to_revision(count, rev, files_touched_in_branch) for count, rev in enumerate(revisions)]

    def changes_in_commit(self) -> set[str]:
        changed: set[str] = set(self._repo.git.diff(self.__get_filter_patterns(), None, name_only=True).splitlines())
        return changed.union(self._repo.untracked_files)

    def changes_in_branch_including_local(self) -> list[Revision]:
        in_branch = self.changes_in_branch()
        in_branch.append(Revision(len(in_branch), self.get_sha, self.changes_in_commit()))
        return in_branch

    def changes_in_tagged_commit(self, current_tag: str) -> list[Revision]:
        curr_rev_tag = self.get_tag

        if curr_rev_tag != current_tag:
            logging.error(f"HEAD is not at {curr_rev_tag} not at expected {current_tag}")
            return []

        return self.changes_in_merge_commit()

    def changes_in_merge_commit(self):
        parent_revs = self._repo.head.commit.parents
        if not parent_revs:
            logging.error("HEAD is not at merge commit, cannot determine changed files.")
            return []
        logging.debug(f"Parent revisions: {parent_revs}")
        files_changed = self._repo.git.diff(f"{str(self._repo.head.commit)}..{str(parent_revs[0])}",
                                            name_only=True).splitlines()
        return [Revision(ord=0, hash=str(self.get_sha), files_touched=files_changed)]

    @property
    def main_branch_pulled(self) -> bool:
        if self._repo.head.is_detached:
            return False
        branch_names = list(map(lambda n: n.name, self._repo.references))
        return f'{self._config.main_branch}' in branch_names

    def __get_remote(self) -> Remote:
        default_remote = self._repo.remote('origin')
        if 'https:' not in default_remote.url or self._config.repo_credentials is None:
            return default_remote

        return default_remote.set_url(self._config.repo_credentials.to_url_with_credentials)

    def pull_main_branch(self):
        remote = self.__get_remote()
        main = self._config.main_branch
        return remote.fetch(f"+refs/heads/{main}:refs/heads/{main}")

    def find_projects(self, folder_pattern: str = '') -> list[str]:
        """ returns a set of all project.yml files
        :type folder_pattern: project paths are filtered on this pattern
        """
        projects = set(self._repo.git.ls_files(f'*{folder_pattern}*/{Project.project_yaml_path()}').splitlines())
        return sorted(projects)



"""A client to connect to an S3 bucket"""

import pathlib
from dataclasses import dataclass
from typing import Optional
from logging import Logger
from os import walk
from http import HTTPStatus
import boto3
from botocore.exceptions import ClientError

from ...project import Project
from ...steps.models import RunProperties


@dataclass
class S3ClientConfig:
    service_name = 's3'
    region_name = 'eu-central-1'
    access_key_id: Optional[str]
    secret_access_key: Optional[str]
    bucket_name: str
    bucket_root_path: str

    def __init__(self, run_properties: RunProperties, bucket_name: str):
        s3_connection_info = run_properties.config.get('s3', None)
        self.access_key_id = s3_connection_info.get('accessKeyId')
        self.secret_access_key = s3_connection_info.get('secretAccessKey')
        self.bucket_name = bucket_name
        self.bucket_root_path = run_properties.versioning.identifier


class S3Client:

    def __init__(self, logger: Logger, config: S3ClientConfig):
        """
        Creates a client that provides a wrapper for uploading a directory to an S3 bucket

        :param config: the S3 client config containing the necessary credentials and bucket information
        """
        self._client = boto3.client(service_name=config.service_name,
                                    region_name=config.region_name,
                                    aws_access_key_id=config.access_key_id,
                                    aws_secret_access_key=config.secret_access_key)
        self._buck_name = config.bucket_name
        self._bucket_root_path = config.bucket_root_path
        self._logger = logger
        self.__validate_bucket()

    def __validate_bucket(self):
        """
        Validates that the bucket exists and that the user has access to it
        """
        try:
            self._client.head_bucket(Bucket=self._buck_name)
            self._logger.info(f"{self._buck_name} is a valid bucket")
        except ClientError as exc:
            error_code = int(exc.response['Error']['Code'])
            if error_code == HTTPStatus.FORBIDDEN:
                self._logger.warning(f"Unable to access bucket {self._buck_name}: Forbidden")
            elif error_code == HTTPStatus.NOT_FOUND:
                self._logger.warning(f"Unable to access bucket {self._buck_name}: Not found")
            raise exc

    def upload_directory(self, directory: str):
        """
        Uploads all files in a directory and its subdirectories to the S3 bucket

        Note: boto3 does not provide a sync method, so we have to walk the directory and upload each file individually

        :param directory: the name of the root directory containing the content to upload
        """
        walks = walk(directory)
        for path, _, filenames in walks:
            for filename in filenames:
                src_path = f'{path}/{filename}'
                dst_path = self.create_file_dst(root_path=self._bucket_root_path, file_path=path, filename=filename)
                self.upload_file(src_path=src_path, dst_path=dst_path)

    @staticmethod
    def create_file_dst(root_path: str, file_path: str, filename: str):
        """
        Creates the bucket key for a given file based on its relative path and the given bucket root path
        i.e. /root/path/relative/path/filename

        The first 2 parts of the path are excluded as they are an auto created tmp directory of the form
        '/tmp/asVx3sd/...'

        :param root_path: the root path on the bucket
        :param file_path: the relative path of the file
        :param filename: the name of the file
        :return: the full destination path/key in the s3 bucket
        """
        path_parts = pathlib.Path(file_path).parts
        non_tmp_parts = path_parts[3:]
        relative_path = f"{'/'.join(filter(None, non_tmp_parts))}"
        return f'{root_path}/{relative_path}/{filename}' if non_tmp_parts else f'{root_path}/{filename}'

    def upload_file(self, src_path: str, dst_path: str):
        """
        Uploads an individual file to the S3 bucket

        :param src_path: the local relative path of the file
        :param dst_path: the destination path within the bucket
        """
        self._logger.debug(f"Uploading to bucket: {self._buck_name}")
        self._logger.debug(f"Uploading file at '{src_path}' to '{dst_path}'")

        try:
            self._client.upload_file(src_path, self._buck_name, dst_path)
        except ClientError as exc:
            self._logger.warning(f'Unexpected exception uploading file: {src_path} - {exc}')
            raise exc



"""
Markdown run result formatters
"""
import operator
import itertools

from junitparser import TestSuite

from ...project import Stage, Project
from ...steps import Output, ArtifactType
from ...steps.models import Artifact
from ...steps.run import RunResult
from ...steps.steps import StepResult
from ...utilities.junit import TestRunSummary, sum_suites, TEST_RESULTS_URL_KEY, TEST_RESULTS_URL_NAME_KEY, \
    to_test_suites


def summary_to_markdown(summary: TestRunSummary):
    return f"🧪 {summary.tests} ❌ {summary.failures} " \
           f"💔 {summary.errors} 🙈 {summary.skipped}"


def __add_link_if_service(name: str, output: Output) -> str:
    if output.produced_artifact and output.produced_artifact.artifact_type == ArtifactType.DEPLOYED_HELM_APP:
        url = output.produced_artifact.spec.get('url')
        if url:
            return f'[{name}]({url})'

    return name


def wrap_project_name(proj: Project, result: list[StepResult]):
    project_name = proj.name
    encapsulation = '_'
    found_result = next((r for r in result if r.project.name == project_name), None)
    if found_result:
        project_name = __add_link_if_service(project_name, found_result.output)
        encapsulation = '*' if found_result.output.success else '~~'

    return f'{encapsulation}{project_name}{encapsulation}'


def __to_oneliner(result: list[StepResult], plan: set[Project]) -> str:
    project_names: list[str] = []
    if plan:
        sorted_plan = sorted(plan, key=operator.attrgetter('name'))
        for proj in sorted_plan:
            project_names.append(wrap_project_name(proj, result))
    else:
        project_names = list(map(lambda r: f'_{r.project.name}_', result))

    return f'{", ".join(project_names)}'


def stage_to_icon(stage: Stage):
    if stage == Stage.BUILD:
        return '🏗️'
    if stage == Stage.TEST:
        return '📋'
    if stage == Stage.DEPLOY:
        return '🚀'
    if stage == Stage.POST_DEPLOY:
        return '🦺'
    return '➡️'


def markdown_for_stage(run_result: RunResult, stage: Stage):
    step_results: list[StepResult] = run_result.results_for_stage(stage)
    plan: set[Project] = run_result.plan_for_stage(stage)
    if not step_results and not plan:
        return ''

    result = f"{stage_to_icon(stage)} {__to_oneliner(step_results, plan)}  \n"
    test_artifacts = _collect_test_artifacts(step_results)
    test_results = _collect_test_results(test_artifacts)

    if test_results:
        result += to_markdown_test_report(test_results)
        unique_artifacts = _collect_unique_test_artifacts_with_url(test_artifacts)

        for unique_artifact in unique_artifacts:
            result += f' [{unique_artifact.spec[TEST_RESULTS_URL_NAME_KEY]}]' \
                      f'({unique_artifact.spec[TEST_RESULTS_URL_KEY]})'

        result += '  \n'

    return result


def run_result_to_markdown(run_result: RunResult) -> str:
    result: str = f'{run_result.status_line}  \n'
    exception = run_result.exception
    if exception:
        result += f"For _{exception.executor}_ on _{exception.project_name}_ at _{exception.stage}_ \n"
        result += f"\n\n{exception}\n\n"
    elif run_result.failed_result:
        failed = run_result.failed_result
        result += f"For _{failed.project.name}_ at _{failed.stage}_ \n"
        result += f"\n\n{run_result.failed_result.output.message}\n\n"

    for stage in Stage:
        result += markdown_for_stage(run_result, stage)

    return result


def to_markdown_test_report(suites: list[TestSuite]):
    total_tests = sum_suites(suites)
    return f"{summary_to_markdown(total_tests)}"


def _collect_test_artifacts(step_results: list[StepResult]) -> list[Artifact]:
    return [res.output.produced_artifact for res in step_results if
            (res.output.produced_artifact and
             res.output.produced_artifact.artifact_type == ArtifactType.JUNIT_TESTS)]


def _collect_test_results(test_artifacts: list[Artifact]) -> list[TestSuite]:
    suites: list[list[TestSuite]] = list(map(to_test_suites, test_artifacts))

    return list(itertools.chain(*suites))


def _collect_unique_test_artifacts_with_url(test_artifacts: list[Artifact]) -> list[Artifact]:
    unique_artifacts: list[Artifact] = []
    for test_artifact in test_artifacts:
        if TEST_RESULTS_URL_KEY in test_artifact.spec and test_artifact.spec[TEST_RESULTS_URL_KEY] != '':
            duplicate_artifact = next((x for x in unique_artifacts if
                                       x.spec[TEST_RESULTS_URL_KEY] == test_artifact.spec[TEST_RESULTS_URL_KEY]), None)
            if not duplicate_artifact:
                test_artifact.spec[TEST_RESULTS_URL_NAME_KEY] = test_artifact.producing_step
                unique_artifacts.append(test_artifact)
            else:
                duplicate_artifact.spec[TEST_RESULTS_URL_NAME_KEY] = 'link'

    return unique_artifacts



"""Jenkins multibranch pipeline build tool"""
import subprocess
from dataclasses import dataclass
from typing import Optional

import requests
from github import Github
from jenkinsapi.jenkins import Jenkins
from rich.console import Console
from rich.markdown import Markdown
from rich.status import Status

from . import play_sound, Sound
from ....project import Target
from ....utilities.github import GithubConfig, get_pr_for_branch
from ....utilities.jenkins import JenkinsConfig, Pipeline
from ....utilities.jenkins.runner import JenkinsRunner
from ....utilities.repo import RepoConfig, Repository


@dataclass(frozen=True)
class JenkinsRunParameters:
    jenkins_user: str
    jenkins_password: str
    config: dict
    pipeline: str
    pipeline_parameters: dict
    verbose: bool
    follow: bool
    tag: Optional[str] = None


def get_token(github_config: GithubConfig):
    if github_config.token:
        return github_config.token
    return subprocess.run(['gh', 'auth', 'token'], stdout=subprocess.PIPE, stderr=subprocess.STDOUT,
                          check=True).stdout.decode('utf-8').strip()


def __get_pr_pipeline(config: dict, git_repo: Repository, pipeline: str, status: Status) -> Optional[Pipeline]:
    github_config = GithubConfig.from_config(config)
    github = Github(login_or_token=get_token(github_config))

    repo = github.get_repo(github_config.repository)

    branch = git_repo.get_branch
    if not branch:
        status.console.log('Could not determine current branch')
        return None

    if git_repo.main_branch == branch:
        status.console.log(f'On main branch ({branch}), cannot determine which PR to build')
        return None
    try:
        pull = get_pr_for_branch(repo, branch)
        return Pipeline(target=Target.PULL_REQUEST, tag=f'{pull.number}', url=pull.html_url, pipeline=pipeline,
                        body=pull.body, jenkins_config=JenkinsConfig.from_config(config))
    except ValueError as exc:
        status.console.log(exc)
        return None


def run_jenkins(run_config: JenkinsRunParameters):
    log_console = Console(log_path=False, log_time=False)
    with log_console.status('Fetching Github info.. [bright_blue]>gh pr view[/bright_blue]') as status:
        config = run_config.config

        with Repository(RepoConfig.from_config(config)) as git_repo:
            try:
                pipeline_info = Pipeline(
                    target=Target.ACCEPTANCE, tag=run_config.tag, url="https://tag-url",
                    pipeline=run_config.pipeline, body="",
                    jenkins_config=JenkinsConfig.from_config(config)) if run_config.tag \
                    else __get_pr_pipeline(config, git_repo, run_config.pipeline, status)
                if not pipeline_info:
                    return

                status.start()
                status.update(f'Fetching Jenkins info for {pipeline_info.human_readable()} ...')

                runner = JenkinsRunner(pipeline=pipeline_info,
                                       jenkins=Jenkins(baseurl=JenkinsConfig.from_config(config).url,
                                                       username=run_config.jenkins_user,
                                                       password=run_config.jenkins_password),
                                       status=status,
                                       follow=run_config.follow,
                                       verbose=run_config.verbose)
                runner.run(run_config.pipeline_parameters)
            except requests.ConnectionError:
                play_sound(Sound.FAILURE)
                status.console.log('⚠️ Could not connect. Are you on VPN?')
            except Exception as exc:
                status.console.print(Markdown(f'Unexpected exception: {exc}'))
                if run_config.verbose:
                    status.console.print_exception()
                raise exc



"""
# Github reporter

## PR comment
Pipeline results can be reported in the form of an update to the PR body or user comment on the pull request,
using the `PullRequestReporter` class.
The mode of update is determined by the `update_stategy` parameter in the constructor.
You are recommended to use a bot account as the authenticated user.

### Installation instructions
 1. Go to your bot user's personal Github profile settings
 2. Under *Developer Settings* -> *Personal access tokens* -> *Fine grained tokens*, click **Generate new token**
 3. Enable `Read` / `Write` access for *Pull requests*
 4. Make sure that this token is present as `GITHUB_TOKEN` env var in your pipeline's runner.

## Checks

The `CommitCheck` reporter reports your build pipeline's result in the form of a
[check](https://docs.github.com/en/rest/checks).
![Pull request check](documentation_images/pr-check.png)

Checks can be referred to from branch protection rules, in order to prevent faulty code from being merged.

### Installation instructions

 1. Install the [https://github.com/apps/mpyl-pipeline](https://github.com/apps/mpyl-pipeline)
 Github app to your repository.
 2. Go to your repository's *Setting* -> *Integrations* -> *Github Apps* and click **Configure** for*_MPyL Pipeline*
 3. In the app's settings page, scroll down and click **Generate a private key**
 4. The private key needs to be made available at the location configured in `csv.targets.app.privateKeyPath` at
  runtime.

"""
import base64
from datetime import datetime
from enum import Enum
from logging import Logger
from pathlib import Path
from typing import Dict, Optional, Callable

from github import Github, GithubIntegration, GithubException
from github.IssueComment import IssueComment
from github.PullRequest import PullRequest
from github.Repository import Repository as GithubRepository

from . import Reporter, ReportOutcome
from ...reporting.formatting.markdown import run_result_to_markdown
from ...reporting.formatting.text import to_string
from ...steps.models import RunProperties
from ...steps.run import RunResult
from ...utilities.github import GithubConfig, get_pr_for_branch, GithubAppConfig
from ...utilities.repo import Repository, RepoConfig


def compose_message_body(results: RunResult, _unused_config: Optional[Dict] = None) -> str:
    return run_result_to_markdown(results)


class GithubOutcome(ReportOutcome):
    pass


class GithubUpdateStategy(Enum):
    BODY = 'body'
    COMMENT = 'comment'


class PullRequestReporter(Reporter):
    _config: GithubConfig

    def __init__(self, config: Dict,
                 compose_function: Callable[[RunResult, Optional[Dict]], str] = compose_message_body,
                 update_stategy: GithubUpdateStategy = GithubUpdateStategy.BODY):
        self._raw_config = config
        self._config = GithubConfig.from_config(config)
        self.git_repository = Repository(RepoConfig.from_config(config))
        self.compose_function = compose_function
        self.update_strategy: GithubUpdateStategy = update_stategy

        self.body_separator = "----"

    def _get_pull_request(self, repo: GithubRepository, run_properties: RunProperties) -> Optional[PullRequest]:
        if run_properties.versioning.pr_number:
            return repo.get_pull(run_properties.versioning.pr_number)

        current_branch = self.git_repository.get_branch
        if current_branch:
            return get_pr_for_branch(repo, current_branch)
        return None

    def _update_pr(self) -> Callable[[PullRequest, RunResult], None]:
        if self.update_strategy == GithubUpdateStategy.COMMENT:
            return self._change_pr_comment
        return self._change_pr_body

    def _extract_pr_header(self, current_body: Optional[str]) -> str:
        body_header = current_body.split(self.body_separator)[0] if current_body else ""
        return body_header.rstrip("\n") + f"\n{self.body_separator}\n"

    def _change_pr_body(self, pull_request: PullRequest, results: RunResult):
        pull_request.edit(
            body=self._extract_pr_header(pull_request.body) + self.compose_function(results, self._raw_config)
        )

    def _change_pr_comment(self, pull_request: PullRequest, results: RunResult):
        github = Github(self._config.token)
        comments = pull_request.get_issue_comments()
        authenticated_user = github.get_user()
        comments_for_user = [c for c in comments if c.user.id == authenticated_user.id]
        if comments_for_user:
            comment_to_update: IssueComment = comments_for_user.pop()
            comment_to_update.edit(self.compose_function(results, self._raw_config))
        else:
            pull_request.create_issue_comment(self.compose_function(results, self._raw_config))

    def send_report(self, results: RunResult, text: Optional[str] = None) -> GithubOutcome:
        try:
            github = Github(self._config.token)
            repo = github.get_repo(self._config.repository)

            pull_request = self._get_pull_request(repo, results.run_properties)
            if not pull_request:
                return GithubOutcome(success=False, exception=Exception('No pull request found'))

            self._update_pr()(pull_request, results)
            return GithubOutcome(success=True)
        except GithubException as exc:
            return GithubOutcome(success=False, exception=exc)


class CommitCheck(Reporter):
    _github_config: GithubConfig
    _check_run_id: Optional[int]

    def __init__(self, config: Dict, logger: Logger):
        self._config = config
        self._github_config = GithubConfig.from_config(config)
        self._check_run_id = None
        self._logger = logger

    @staticmethod
    def _to_output(results: RunResult) -> dict:
        build_id = results.run_properties.details.build_id
        summary = ':white_check_mark: Build successful' if results.is_success else ':x: Build failed'
        return {'title': f'Build {build_id}', 'summary': summary + '\n' + run_result_to_markdown(results),
                'text': to_string(results)}

    def send_report(self, results: RunResult, text: Optional[str] = None) -> GithubOutcome:
        try:
            config: GithubAppConfig = self._github_config.get_app_config
            if not config:
                raise KeyError("github.app config needs to be defined")

            private_key = Path(config.private_app_key_path or '').read_text(
                encoding='utf-8') if config.private_app_key_path else base64.b64decode(
                config.private_key_base_64_encoded or '').decode('utf-8')

            integration = GithubIntegration(integration_id=config.app_key, private_key=private_key)

            install = integration.get_installation(self._github_config.owner, self._github_config.repo_name)
            access_token = integration.get_access_token(install.id)
            github = Github(login_or_token=access_token.token)
            repo = github.get_repo(self._github_config.repository)
            if self._check_run_id and results:
                run = repo.get_check_run(self._check_run_id)
                conclusion = 'success' if results is None or results.is_success else 'failure'
                self._logger.info(f'Setting check to {conclusion}')
                run.edit(completed_at=datetime.now(), conclusion=conclusion, output=self._to_output(results))
            else:
                with Repository(RepoConfig.from_config(self._config)) as git_repository:
                    self._check_run_id = repo.create_check_run(name='Pipeline build', head_sha=git_repository.get_sha,
                                                               status='in_progress').id
            return GithubOutcome(success=True)
        except GithubException as exc:
            self._logger.warning(f'Unexpected exception: {exc}', exc_info=True)
            return GithubOutcome(success=False, exception=exc)



"""
.. include:: ../../README.md

.. include:: ../../README-usage.md

.. include:: ../../README-dev.md
"""

import click

from .cli.build import build
from .cli.health import health
from .cli.meta_info import get_version
from .cli.meta_info import version
from .cli.projects import projects
from .utilities.pyaml_env import parse_config
from .utilities.repo import RepoConfig, Repository


@click.group(name='mpyl')
def main_group():
    """Command Line Interface for MPyL"""


def main():
    main_group.help = f"Command Line Interface for MPyL {get_version()}"
    main_group.add_command(projects)
    main_group.add_command(build)
    main_group.add_command(version)
    main_group.add_command(health)
    main_group()  # pylint: disable = no-value-for-parameter



"""
Interface for custom reporters.
To create a new reporter, implement `Reporter.send_report`.
You can transform `mpyl.steps.run.RunResult` to a report using any of the standard formatters under
`mpyl.reporting.formatting` and send it to target of your choice.

See the **Submodules** for built-in reporters.
"""
from abc import abstractmethod
from dataclasses import dataclass
from typing import Optional

from ...steps.run import RunResult


@dataclass(frozen=True)
class ReportOutcome:
    success: bool
    exception: Optional[Exception] = None


class ReportAccumulator:
    outcomes: list[ReportOutcome]

    def __init__(self):
        self.outcomes = []

    def add(self, outcome: ReportOutcome):
        self.outcomes.append(outcome)

    @property
    def failures(self) -> list[str]:
        return [f'{type(outcome).__name__} failed with exception {outcome.exception}' for outcome in self.outcomes if
                not outcome.success]


class Reporter:

    @abstractmethod
    def send_report(self, results: RunResult, text: Optional[str] = None) -> ReportOutcome:
        pass



"""CLI support for Jenkins multibranch pipeline"""
import os
import shutil
from dataclasses import dataclass
from enum import Enum

from rich.console import Console


@dataclass(frozen=True)
class Sound(Enum):
    SUCCESS = 'Glass.aiff'
    FAILURE = 'Sosumi.aiff'


def play_sound(sound: Sound):
    if shutil.which('afplay') is None:
        Console().bell()
        return

    os.system(f'afplay /System/Library/Sounds/{sound.value}')



"""Health checks"""

import asyncio
import os
import pkgutil
import shutil
from pathlib import Path
from subprocess import CalledProcessError

import jsonschema
from dotenv import load_dotenv
from rich.console import Console
from rich.markdown import Markdown

from ..build.jenkins import get_token
from ....constants import DEFAULT_CONFIG_FILE_NAME, DEFAULT_RUN_PROPERTIES_FILE_NAME
from ....utilities.github import GithubConfig
from ....utilities.jenkins import JenkinsConfig
from ....utilities.pyaml_env import parse_config
from ....cli import fetch_latest_version, get_meta_version
from ....validation import validate


class HealthConsole:
    def __init__(self, console: Console):
        self.console = console

    def title(self, title: str):
        self.console.print(Markdown(f'*{title}*'))

    def check(self, check: str, success: bool):
        icon = '✅' if success else '❌'
        self.console.print(Markdown(f'&nbsp;&nbsp;{icon} {check}'))


def perform_health_checks(bare_console: Console):
    console = HealthConsole(bare_console)
    console.title('Version')
    __check_version(console)

    console.title('MPyL configuration')
    __check_config(console, env_var='MPYL_CONFIG_PATH', default=DEFAULT_CONFIG_FILE_NAME,
                   schema_path='../../../schema/mpyl_config.schema.yml', name='config')

    console.title('Run configuration')
    __check_config(console, env_var='MPYL_RUN_PROPERTIES_PATH', default=DEFAULT_RUN_PROPERTIES_FILE_NAME,
                   schema_path='../../../schema/run_properties.schema.yml', name='run properties')

    console.title('Jenkins')
    __check_jenkins(console)


def __check_jenkins(console: HealthConsole):
    path = os.environ.get('MPYL_CONFIG_PATH', default=DEFAULT_CONFIG_FILE_NAME)
    if not os.path.exists(path):
        console.check(f'Configuration not found at: `{path}`', success=False)
        return

    parsed = parse_config(Path(path))

    try:
        jenkins_conf = JenkinsConfig.from_config(parsed)
        console.check(f'Jenkins configured for pipeline `{jenkins_conf.default_pipeline}` '
                      f'at [{jenkins_conf.url}]({jenkins_conf.url})', success=True)
    except KeyError as exc:
        console.check(f'Jenkins config not valid: {exc}', success=False)

    gh_is_installed = shutil.which('gh')
    if gh_is_installed:
        console.check('Github cli client `gh` installed', success=True)
    else:
        console.check(
            'Github cli client `gh` not found. Install via [https://cli.github.com/](https://cli.github.com/) '
            'and run `gh auth login`', success=False)

    if gh_is_installed:
        try:
            get_token(GithubConfig.from_config(parsed))
            console.check('Github token found', success=True)
        except CalledProcessError:
            console.check('Github token not found. Log in with `gh auth login`', success=False)

    if os.environ.get('JENKINS_USER'):
        console.check('Jenkins user set', success=True)
    else:
        jenkins_url = f'{JenkinsConfig.from_config(parsed).url}user/me@vandebron.nl/configure'
        message = f'Jenkins user not set via JENKINS_USER env var. Create a user API token in Jenkins' \
                  f' (user:password) API token: {jenkins_url}'
        console.check(message, success=False)

    if os.environ.get('JENKINS_PASSWORD'):
        console.check('Jenkins password set', success=True)
    else:
        console.check('Jenkins password not set via JENKINS_PASSWORD env var', success=False)


def __check_version(console):
    update = asyncio.get_event_loop().run_until_complete(fetch_latest_version())
    meta_version = get_meta_version()
    if update and meta_version:
        if meta_version == update:
            console.check(f'At latest version: {update}', success=True)
        else:
            console.check(f'Outdated version: {meta_version} (latest: {update})', success=False)
    else:
        console.check('Could not determine latest version', success=False)


def __check_config(console, env_var, default, schema_path, name):
    path = os.environ.get(env_var, default=default)
    location = f"{name} at `/{path}` via environment variable `{env_var}`" if os.environ.get(
        env_var) else f"{name} at `/{path}`"
    if os.path.exists(path):
        console.check(f"Found {location}", success=True)

        if load_dotenv(Path(".env")):
            console.check("Set env variables via .env file", success=True)

        parsed = parse_config(path)
        schema_dict = pkgutil.get_data(__name__, schema_path)
        if schema_dict:
            try:
                validate(parsed, schema_dict.decode('utf-8'))
                console.check(f'{name.capitalize()} is valid', success=True)
            except jsonschema.exceptions.ValidationError as exc:
                console.check(
                    f"{name.capitalize()} is invalid: {exc.message} at '{'.'.join(exc.path)}'. 🤔Did you rebase"
                    f" your branch onto {parsed['cvs']['git']['mainBranch']}?",
                    success=False)
    else:
        console.check(f"Could not find {location}. Location can be specified with env var '{env_var}'", success=False)



"""Configuration required for running cypress"""
from dataclasses import dataclass
from typing import Optional


@dataclass(frozen=True)
class CypressConfig:
    cypress_source_code_path: str
    kubectl_config_path: str
    record_key: Optional[str]

    @staticmethod
    def from_config(config: dict):
        cypress_config = config.get('cypress')
        if not cypress_config:
            raise KeyError('Cypress section needs to be defined in mpyl_config.yml')

        return CypressConfig(cypress_source_code_path=cypress_config.get('cypressSourceCodePath'),
                             record_key=cypress_config.get('recordKey'),
                             kubectl_config_path=cypress_config.get('kubectlConfigPath', '~/.kube/config'))



""" Step that uploads static files to s3 and then deploys the docker image produced in the build stage to Kubernetes,
using HELM."""
from logging import Logger
import tempfile

from .kubernetes import DeployKubernetes
from .. import Step, Meta
from ...project import Stage
from ..models import Input, Output, ArtifactType
from ...utilities.docker import docker_image_tag, docker_copy, create_container
from ...utilities.s3 import S3Client, S3ClientConfig

STATIC_FOLDER = 'static'


class CloudFrontKubernetesDeploy(Step):

    def __init__(self, logger: Logger) -> None:
        super().__init__(logger, Meta(
            name='CloudFront Kubernetes Deploy',
            description='uploads the build output to an s3 bucket',
            version='0.0.1',
            stage=Stage.DEPLOY,
        ), produced_artifact=ArtifactType.NONE, required_artifact=ArtifactType.DOCKER_IMAGE)

    def execute(self, step_input: Input) -> Output:
        with tempfile.TemporaryDirectory() as tmp_folder:
            self.__copy_docker_assets(logger=self._logger, step_input=step_input, tmp_folder=tmp_folder)

            if not step_input.dry_run:
                self.__upload_to_s3(logger=self._logger, step_input=step_input, tmp_folder=tmp_folder)

        return DeployKubernetes(self._logger).execute(step_input)

    @staticmethod
    def __copy_docker_assets(logger: Logger, step_input: Input, tmp_folder: str):
        """
        Copies the static assets from the docker image to a temp folder
        """
        image_name = docker_image_tag(step_input)
        container_path = f'{step_input.project.name}/{STATIC_FOLDER}'
        container = create_container(logger, image_name)
        docker_copy(logger=logger, container_path=container_path, dst_path=tmp_folder, container=container)

    @staticmethod
    def __upload_to_s3(logger: Logger, step_input: Input, tmp_folder: str):
        """
        Creates an S3 client and uploads the static assets stored in the temp folder
        """
        logger.info("Creating S3 client")
        bucket_name = step_input.project.s3_bucket.bucket.get_value(step_input.run_properties.target)
        s3_config = S3ClientConfig(run_properties=step_input.run_properties, bucket_name=bucket_name)
        s3_client = S3Client(logger, s3_config)

        logger.info(f"Uploading assets to '{s3_config.bucket_root_path}' in bucket '{s3_config.bucket_name}'")
        s3_client.upload_directory(directory=tmp_folder)
        logger.info('Upload complete')



""" Step that deploys the docker image produced in the build stage to Kubernetes, using HELM. """
import re
from logging import Logger
from typing import Optional

from .k8s import deploy_helm_chart, CustomResourceDefinition, cluster_config
from .k8s.chart import ChartBuilder, to_service_chart
from .. import Step, Meta
from ..models import Input, Output, ArtifactType, input_to_artifact
from ...project import Stage, Target
from ...stages.discovery import find_deploy_set
from ...utilities.repo import RepoConfig

DEPLOYED_SERVICE_KEY = 'url'


class DeployKubernetes(Step):

    def __init__(self, logger: Logger) -> None:
        super().__init__(logger, Meta(
            name='Kubernetes Deploy',
            description='Deploy to k8s',
            version='0.0.1',
            stage=Stage.DEPLOY
        ), produced_artifact=ArtifactType.NONE, required_artifact=ArtifactType.DOCKER_IMAGE)

    @staticmethod
    def match_to_url(match: str) -> str:
        return 'https://' + next(iter(re.findall(r'`(.*)`', match.split(',')[-1])))

    @staticmethod
    def try_extract_hostname(chart: dict[str, CustomResourceDefinition]) -> Optional[str]:
        ingress = chart.get('ingress-https-route')
        if ingress:
            routes = ingress.spec.get('routes', {})
            if routes:
                match = routes[0].get('match')
                return DeployKubernetes.match_to_url(match)
        return None

    def execute(self, step_input: Input) -> Output:
        properties = step_input.run_properties
        builder = ChartBuilder(step_input, find_deploy_set(repo_config=RepoConfig.from_config(properties.config),
                                                           tag=step_input.run_properties.versioning.tag))
        chart = to_service_chart(builder)

        target_cluster = cluster_config(properties.target, properties)
        deploy_result = deploy_helm_chart(self._logger, chart, step_input, target_cluster, builder.release_name)
        if deploy_result.success:
            hostname = self.try_extract_hostname(chart)
            spec = {}
            if hostname:
                has_specific_routes_configured: bool = bool(builder.deployment.traefik is not None)
                self._logger.info(f"Service {step_input.project.name} reachable at: {hostname}")

                endpoint = '/' if has_specific_routes_configured else '/swagger/index.html'
                spec[DEPLOYED_SERVICE_KEY] = f'{hostname}{endpoint}'
            artifact = input_to_artifact(ArtifactType.DEPLOYED_HELM_APP, step_input, spec=spec)
            if properties.target == Target.PRODUCTION:
                self._logger.info(
                    f"Release to production successful, updating base images in {Target.PULL_REQUEST_BASE} "
                    f"to make sure the Test environment is in sync with production")
                target_cluster = cluster_config(Target.PRODUCTION, properties)
                deploy_to_prod_result = deploy_helm_chart(self._logger, chart, step_input, target_cluster,
                                                          builder.release_name)
                return Output(success=True, message=deploy_result.message + '\n' + deploy_to_prod_result.message,
                              produced_artifact=artifact)

            return Output(success=True, message=deploy_result.message, produced_artifact=artifact)

        return deploy_result
